__author__ = 'Luca Rebuffi'

"""
=========
Loop Management
=========

Widgets for loop management

"""

# Category description for the widget registry

#NAME = "OASYS \u2023 Basic Loops"
NAME = "OASYS \u23F5 Loops"

ID = "oasys2.widgets.loops"

DESCRIPTION = "Widgets for Loop Management."

BACKGROUND = "#FAAC58"

ICON = "icons/loops.png"

PRIORITY = 1.1
